(* Theorems about CPS conversion *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".
Require Import SourceLang.
Require Import CpsLang.
Require Import Cps.


Lemma cps_ctxt_preserves_in :
  forall (C : sctxt) (x : var) (T : stype),
  in' x T C ->
  in' x (cps_type T) (cps_ctxt C).
Proof.
  induction C as [ | [ x' T' ] C' IH ];
  intros x T H.
    case H.

    destruct H as [ [ <- <- ] | [ H1 H2 ] ]; fold (@in' var stype) in *.
      left; split; reflexivity.

      right; split.
        exact H1.

        apply IH, H2.
Qed.

Theorem cps_term_preserves_typing :
  forall (C : sctxt) (T : stype) (t : sterm),
  sterm_typing C t T ->
  forall Cc : cctxt,
  cterm_typing (cps_ctxt C) Cc (cps_term t) (! ! cps_type T).
Proof.
  induction 1 as [ C x T H1
                 | C
                 | C x TA TR t H1 IH1
                 | C TA TR tf ta H1 IH1 H2 IH2 ];
  intro Cc;
  simpl.

  (* case stt_fvar *)
  apply ctt_cabs.
  apply ctt_app with (TA := cps_type T).
    apply ctt_fcvar.
    left; split; reflexivity.

    apply ctt_fsvar.
    apply cps_ctxt_preserves_in, H1.

  (* case stt_unit *)
  apply ctt_cabs.
  apply ctt_app with (TA := cps_type sy_unit).
    apply ctt_fcvar.
    left; split; reflexivity.

    apply ctt_unit.

  (* case stt_abs *)
  apply ctt_cabs.
  apply ctt_app with (TA := cps_type (sy_func TA TR)).
    apply ctt_fcvar.
    left; split; reflexivity.

    apply ctt_sabs; fold cps_type.
    apply IH1.

  (* case stt_app *)
  apply ctt_cabs.
  apply ctt_app with (TA := ! cps_type (sy_func TA TR)).
    apply IH1.

    apply ctt_cabs.
    apply ctt_app with (TA := ! cps_type TA).
      apply IH2.

      apply ctt_cabs.
      apply ctt_app with (TA := ! cps_type TR);
      [ apply ctt_app with (TA := cps_type TA) | ];
      apply ctt_fcvar;
      repeat ((left; split; reflexivity)
          || (right; split; [ discriminate | ])).
Qed.
