(* Definition of the source language *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".


Section Terms.

(* term *)
Inductive sterm : Set :=
  | st_fvar (x : var)
  | st_unit
  | st_abs (x : var) (t : sterm)
  | st_app (tf ta : sterm).

Fixpoint fvars_in_sterm (t : sterm) : list var :=
  match t with
  | st_fvar x =>
      x :: nil
  | st_unit =>
      nil
  | st_abs x t1 =>
      remove eq_var_dec x (fvars_in_sterm t1)
  | st_app tf ta =>
      fvars_in_sterm tf ++ fvars_in_sterm ta
  end.

End Terms.


Section Types.

(* type *)
Inductive stype : Set :=
  | sy_unit
  | sy_func (TA TR : stype).

End Types.


Section Contexts.

(* typing context *)
Definition sctxt : Set :=
  list (var * stype).

Definition domain_of_sctxt : sctxt -> list var :=
  map (@fst _ _).

End Contexts.


Section Rules.

(* term typing rule *)
Inductive sterm_typing : sctxt -> sterm -> stype -> Prop :=
  | stt_fvar :
      forall (C : sctxt) (x : var) (T : stype),
      in' x T C ->
      sterm_typing C (st_fvar x) T
  | stt_unit :
      forall C : sctxt,
      sterm_typing C st_unit sy_unit
  | stt_abs :
      forall (C : sctxt) (x : var) (TA TR : stype) (t : sterm),
      sterm_typing ((x, TA) :: C) t TR ->
      sterm_typing C (st_abs x t) (sy_func TA TR)
  | stt_app :
      forall (C : sctxt) (TA TR : stype) (tf ta : sterm),
      sterm_typing C tf (sy_func TA TR) ->
      sterm_typing C ta TA ->
      sterm_typing C (st_app tf ta) TR.

End Rules.
